#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>

#define MAX_LEN_TUPLE 18
#define MAX_LEN_FREQ   3
#define NUM_STRAND     1
#define MAX_NUM_WORD   4

#define NUM_NUCLEOTIDE 4

#define min(a,b) ((a < b) ? a : b)
#define max(a,b) ((a > b) ? a : b)
#define nucleotide(c) ((c=='a')?0:((c=='c')?1:((c=='g')?2:((c=='t')?3:-1))))

typedef struct freq_node_type freq_node;
struct freq_node_type
{
   int *freq, num_next;
   freq_node **next;
};

typedef struct word_node_type word_node;
struct word_node_type
{
   int pos, *num_word;
   word_node *next [NUM_NUCLEOTIDE];
};

typedef struct motif_node_type motif_node;
struct motif_node_type
{
   int len_tuple, num_word;
   long double prob;
   char *str;
   motif_node *next;
};

char nucleotide_char [] =
{
   /* 0 */ 'a', /* 1 */ 'c', /* 2 */ 'g', /* 3 */ 't'
};

char *fmakeword(FILE *f, char stop, int *cl) {
    int wsize;
    char *word;
    int ll;

    wsize = 102400;
    ll=0;
    word = (char *) malloc(sizeof(char) * (wsize + 2));

    while(1) {
        word[ll] = (char)fgetc(f);
        if(ll==wsize) {
            word[ll+1] = '\0';
            wsize+=102400;
            word = (char *)realloc(word,sizeof(char)*(wsize+2));
        }
        if (cl != NULL)
           --(*cl);
        if((word[ll] == stop) || (feof(f)) || ((cl != NULL) && (!(*cl)))) {
            if((word[ll] != stop) && (!feof(f)) && ((cl == NULL) || (!(*cl))))
                ll++;
            word[ll] = '\0';
            return word;
        }
        ++ll;
    }
}

char complement (char ch)
{
   if (ch == 'a')
      return 't';
   if (ch == 'A')
      return 'T';
   if (ch == 'c')
      return 'g';
   if (ch == 'C')
      return 'G';
   if (ch == 'g')
      return 'c';
   if (ch == 'G')
      return 'C';
   if (ch == 't')
      return 'a';
   if (ch == 'T')
      return 'A';
   return ch;
}

char **read_fasta (char *str, char ***sequence_name, int *num_sequence)
{
   int i, n, start, end, len_sequence;
   char **sequence;

   n = strlen (str);
   (*num_sequence) = 0;
   i = 0;
   while (i < n)
   {
      while ((i < n) && (str [i] != '>'))
         i++;
      while ((i < n) && (str [i] != '\n'))
         i++;
      while ( (i < n) && (str [i] != '>') && (!islower (str [i]))
           && (!isupper (str [i])) )
         i++;
      if ((i < n) && (str [i] != '>'))
         (*num_sequence)++;
   }
   sequence = (char **) malloc ((*num_sequence) * sizeof (char *));
   if (errno == ENOMEM)
   {
      printf ("Error: Not enough memory for sequence\n");
      exit (-1);
   }
   (*sequence_name) = (char **) malloc ((*num_sequence) * sizeof (char *));
   if (errno == ENOMEM)
   {
      printf ("Error: Not enough memory for sequence_name\n");
      exit (-1);
   }
   (*num_sequence) = 0;
   i = 0;
   while ((i < n) && (str [i] != '>'))
      i++;
   while (i < n)
   {
      start = i + 1;
      while ((i < n) && (str [i] != '\n'))
         i++;
      end = i - 1;
      while ( (i < n) && (str [i] != '>') && (!islower (str [i]))
           && (!isupper (str [i])) )
         i++;
      if ((i < n) && (str [i] != '>'))
      {
         (*sequence_name) [*num_sequence] = (char *)
            malloc ((end - start + 2) * sizeof (char));
         if (errno == ENOMEM)
         {
            printf ("Error: Not enough memory for sequence_name\n");
            exit (-1);
         }
         strncpy ((*sequence_name) [*num_sequence], &str [start],
            end - start + 1);
         (*sequence_name) [*num_sequence][end - start + 1] = '\0';
         start = i;
         len_sequence = 0;
         while ((i < n) && (str [i] != '>'))
         {
            if ((islower (str [i])) || (isupper (str [i])))
               len_sequence++;
            i++;
         }
         sequence [*num_sequence] = (char *)
            malloc ((len_sequence + 1) * sizeof (char));
         if (errno == ENOMEM)
         {
            printf ("Error: Not enough memory for sequence\n");
            exit (-1);
         }
         i = start;
         len_sequence = 0;
         while ((i < n) && (str [i] != '>'))
         {
            if ((islower (str [i])) || (isupper (str [i])))
            {
               sequence [*num_sequence][len_sequence] = str [i];
               len_sequence++;
            }
            i++;
         }
         sequence [*num_sequence][len_sequence] = '\0';
         (*num_sequence)++;
      }
   }
   return sequence;
}

long double get_prob (long double prob, int power)
{
   long double prob2;

   if (power == 0)
      return 1.0;
   prob2 = get_prob (prob, power / 2);
   prob2 *= prob2;
   if (power % 2 > 0)
      prob2 *= prob;
   return prob2;
}

void compute_motif0 (int max_len_freq, int max_num_word, int num_sequence,
   int len_tuple, int len_word, int len_word2, int *pos, int **freq,
   int num_tuple, word_node *word, char *str, int index, int index2,
   long double prob, motif_node **motif)
{
   int i, j, *num_word, num_word2, index3, index4, freq2;
   long double prob2, prob3, prob4, prob5, prob6, prob7;
   motif_node *current_motif;

   if (len_word == len_word2)
   {
      num_word = (int *)
         malloc (min (num_tuple / num_sequence, (max_num_word - 1)
            / num_sequence + 1) * sizeof (int));
      if (errno == ENOMEM)
      {
         printf ("Error: Not enough memory for num_word\n");
         exit (-1);
      }
      for (i = 0; i < min (num_tuple / num_sequence, (max_num_word - 1)
              / num_sequence + 1); i++)
         num_word [i] = 0;
      for (i = 0; i < num_sequence; i++)
      {
         if (word -> num_word [i] > 0)
            num_word [min (min (num_tuple / num_sequence, (max_num_word - 1)
               / num_sequence + 1), word -> num_word [i]) - 1]++;
      }
      free (word -> num_word);
      for (i = min (num_tuple / num_sequence, (max_num_word - 1) / num_sequence
              + 1) - 2; i >= 0; i--)
         num_word [i] += num_word [i + 1];
      prob2 = 1.0;
      prob3 = get_prob (1.0 - prob, num_tuple / num_sequence);
      prob4 = 1.0;
      prob5 = 0.0;
      for (i = 0; i < min (num_tuple / num_sequence, (max_num_word - 1)
              / num_sequence + 1); i++)
      {
         prob5 += prob4;
         prob4 *= 1.0 * (num_tuple / num_sequence - i) / (i + 1) * prob
            / (1.0 - prob);
         if ( (i == min (num_tuple / num_sequence, (max_num_word - 1)
                  / num_sequence + 1) - 1)
           || (num_word [i] > num_word [i + 1]) )
         {
            prob6 = 1.0;
            prob7 = 0.0;
            for (j = 0; j < num_word [i]; j++)
            {
               prob7 += prob6;
               prob6 *= 1.0 * (num_sequence - j) / (j + 1) * (1 - prob3 * prob5)
                  / prob3 / prob5;
            }
            prob7 = 1.0 - get_prob (prob3 * prob5, num_sequence) * prob7;
            for (j = 0; j < len_word; j++)
            {
               if (j < 2)
                  prob7 *= 1.0 * NUM_NUCLEOTIDE;
               else
                  prob7 *= 1.0 * (len_tuple - j) / (j - 1) * NUM_NUCLEOTIDE;
            }
            if ((prob7 > 0.0) && (prob7 < prob2))
            {
               num_word2 = i + 1;
               prob2 = prob7;
            }
         }
      }
      free (num_word);
      if (prob2 < 1.0)
      {
         current_motif = (motif_node *) malloc (sizeof (motif_node));
         if (errno == ENOMEM)
         {
            printf ("Error: Not enough memory for current_motif\n");
            exit (-1);
         }
         current_motif -> len_tuple = len_tuple;
         current_motif -> num_word = num_word2;
         current_motif -> prob = prob2;
         current_motif -> str = (char *) malloc (len_tuple * sizeof (char));
         if (errno == ENOMEM)
         {
            printf ("Error: Not enough memory for str\n");
            exit (-1);
         }
         strncpy (current_motif -> str, str, len_tuple);
         current_motif -> next = (*motif);
         (*motif) = current_motif;
      }
   }
   else
   {
      for (i = 0; i < NUM_NUCLEOTIDE; i++)
      {
         if (word -> next [i] != NULL)
         {
            str [pos [len_word2]] = nucleotide_char [i];
            index3 = index;
            if (len_word2 < max_len_freq)
               index3 *= NUM_NUCLEOTIDE;
            index4 = index2;
            if (len_word2 >= max_len_freq)
               index4 %= index / NUM_NUCLEOTIDE;
            index4 = NUM_NUCLEOTIDE * index4 + i;
            prob2 = prob;
            if (len_word2 >= max_len_freq)
            {
               freq2 = 0;
               for (j = 0; j < NUM_NUCLEOTIDE; j++)
                  freq2 += freq [len_word2 - max_len_freq + 1][index4
                     / NUM_NUCLEOTIDE * NUM_NUCLEOTIDE + j];
               prob2
                  *= 1.0 * freq [len_word2 - max_len_freq + 1][index4] / freq2;
            }
            else if ( (len_word == len_word2 + 1)
                   || (len_word2 + 1 == max_len_freq) )
               prob2 *= 1.0 * freq [0][index4] / freq [0][index3];
            compute_motif0 (max_len_freq, max_num_word, num_sequence, len_tuple,
               len_word, len_word2 + 1, pos, freq, num_tuple, word -> next [i],
               str, index3, index4, prob2, motif);
         }
      }
   }
   free (word);
}

void compute_motif (int max_len_freq, int num_strand, int max_num_word,
   int num_sequence, char **sequence, freq_node **freq, freq_node *current_freq,
   FILE *file, int len_tuple, int len_word, int len_word2, int *pos, int pos2,
   motif_node **motif)
{
   int i, j, k, l, m, **freq2, index, num_tuple, len_sequence, found;
   char ch, *str;
   freq_node *current_freq2;
   word_node *word, *current_word;

   if (len_word <= len_word2 + 1)
   {
      freq2 = (int **)
         malloc ((max (0, len_word - max_len_freq) + 1) * sizeof (int *));
      if (errno == ENOMEM)
      {
         printf ("Error: Not enough memory for freq2\n");
         exit (-1);
      }
      if (len_word <= max_len_freq)
      {
         index = 1;
         for (i = 0; i < len_word; i++)
            index *= NUM_NUCLEOTIDE;
         freq2 [0] = (int *) malloc ((index + 1) * sizeof (int));
         if (errno == ENOMEM)
         {
            printf ("Error: Not enough memory for freq2\n");
            exit (-1);
         }
         freq2 [0][index] = 0;
         for (i = 0; i < index; i++)
         {
            if (fscanf (file, "%d", &freq2 [0][i]) == EOF)
            {
               printf ("Error: Unexpected end of file\n");
               exit (-1);
            }
            freq2 [0][index] += freq2 [0][i];
         }
      }
      else
      {
         for (i = 0; i <= len_word - max_len_freq; i++)
         {
            current_freq2 = freq [pos [i + max_len_freq - 1] - pos [i]
               - max_len_freq + 1];
            for (j = i; j < i + max_len_freq - 2; j++)
               current_freq2
                  = current_freq2 -> next [pos [j + 1] - pos [j] - 1];
            freq2 [i] = current_freq2 -> freq;
         }
      }
      num_tuple = 0;
      for (i = 0; i < num_sequence; i++)
         num_tuple += max (0, (int) (strlen (sequence [i])) - len_tuple + 1)
            * num_strand;
      word = (word_node *) malloc (sizeof (word_node));
      if (errno == ENOMEM)
      {
         printf ("Error: Not enough memory for word\n");
         exit (-1);
      }
      for (i = 0; i < NUM_NUCLEOTIDE; i++)
         word -> next [i] = NULL;
      for (i = 0; i < num_sequence; i++)
      {
         len_sequence = strlen (sequence [i]);
         for (j = 0; j < len_sequence - len_tuple + 1; j++)
         {
            for (k = 0; k < num_strand; k++)
            {
               found = 0;
               for (l = 0; l < len_word; l++)
               {
                  if (k == 0)
                     ch = sequence [i][j + pos [l]];
                  else
                     ch = complement (sequence [i][j + len_tuple - pos [l]
                        - 1]);
                  if (nucleotide (tolower (ch)) < 0)
                  {
                     found = 1;
                     break;
                  }
               }
               if (!found)
               {
                  current_word = word;
                  for (l = 0; l < len_word; l++)
                  {
                     if (k == 0)
                        ch = sequence [i][j + pos [l]];
                     else
                        ch = complement (sequence [i][j + len_tuple - pos [l]
                           - 1]);
                     if (current_word -> next [nucleotide (tolower (ch))]
                            == NULL)
                     {
                        current_word -> next [nucleotide (tolower (ch))]
                           = (word_node *) malloc (sizeof (word_node));
                        if (errno == ENOMEM)
                        {
                           printf ("Error: Not enough memory for next\n");
                           exit (-1);
                        }
                        if (l < len_word - 1)
                        {
                           for (m = 0; m < NUM_NUCLEOTIDE; m++)
                              current_word -> next [nucleotide (tolower (ch))]
                                 -> next [m] = NULL;
                        }
                        else
                        {
                           current_word -> next [nucleotide (tolower (ch))]
                              -> num_word
                              = (int *) malloc (num_sequence * sizeof (int));
                           if (errno == ENOMEM)
                           {
                              printf
                                 ("Error: Not enough memory for num_word\n");
                              exit (-1);
                           }
                           for (m = 0; m < num_sequence; m++)
                              current_word -> next [nucleotide (tolower (ch))]
                                 -> num_word [m] = 0;
                        }
                     }
                     current_word = current_word
                        -> next [nucleotide (tolower (ch))];
                  }
                  if ( (current_word -> num_word [i] == 0)
                    || (j >= current_word -> pos + len_tuple) )
                  {
                     current_word -> pos = j;
                     current_word -> num_word [i]++;
                  }
               }
            }
         }
      }
      str = (char *) malloc (len_tuple * sizeof (char));
      if (errno == ENOMEM)
      {
         printf ("Error: Not enough memory for str\n");
         exit (-1);
      }
      for (i = 0; i < len_tuple; i++)
         str [i] = ' ';
      compute_motif0 (max_len_freq, max_num_word, num_sequence, len_tuple,
         len_word, 0, pos, freq2, num_tuple, word, str, 1, 0, 1.0, motif);
      free (str);
      if (len_word <= max_len_freq)
      {
         if (current_freq == NULL)
            free (freq2 [0]);
         else
            current_freq -> freq = freq2 [0];
      }
      free (freq2);
   }
   else
   {
      if (current_freq != NULL)
      {
         current_freq -> num_next = len_tuple - len_word + len_word2 - pos2 + 1;
         current_freq -> next = (freq_node **)
            malloc ((len_tuple - len_word + len_word2 - pos2 + 1)
               * sizeof (freq_node *));
         if (errno == ENOMEM)
         {
            printf ("Error: Not enough memory for next\n");
            exit (-1);
         }
      }
      for (i = pos2; i <= len_tuple - len_word + len_word2; i++)
      {
         if (current_freq == NULL)
            current_freq2 = NULL;
         else
         {
            current_freq2 = (freq_node *) malloc (sizeof (freq_node));
            if (errno == ENOMEM)
            {
               printf ("Error: Not enough memory for current_freq2\n");
               exit (-1);
            }
            current_freq2 -> next = NULL;
            current_freq -> next [i - pos2] = current_freq2;
         }
         pos [len_word2] = i;
         compute_motif (max_len_freq, num_strand, max_num_word, num_sequence,
            sequence, freq, current_freq2, file, len_tuple, len_word,
            len_word2 + 1, pos, i + 1, motif);
      }
   }
}

void free_freq (freq_node *freq)
{
   int i;

   if (freq -> next == NULL)
      free (freq -> freq);
   else
   {
      for (i = 0; i < freq -> num_next; i++)
         free_freq (freq -> next [i]);
      free (freq -> next);
   }
   free (freq);
}

int motif_compare (const void *motif, const void *motif2)
{
   if ((*((motif_node **) motif)) -> prob
         < (*((motif_node **) motif2)) -> prob)
      return -1;
   if ((*((motif_node **) motif)) -> prob
         > (*((motif_node **) motif2)) -> prob)
      return 1;
   return 0;
}

void print_motif (int num_strand, int num_sequence, char **sequence,
   char **sequence_name, motif_node *motif)
{
   int i, j, k, l, m, num_motif, **paint, len_sequence, found, found2;
   char ch;
   motif_node *current_motif, **motif2;

   num_motif = 0;
   current_motif = motif;
   while (current_motif != NULL)
   {
      num_motif++;
      current_motif = current_motif -> next;
   }
   motif2 = (motif_node **) malloc (num_motif * sizeof (motif_node *));
   if (errno == ENOMEM)
   {
      printf ("Error: Not enough memory for motif2\n");
      exit (-1);
   }
   num_motif = 0;
   current_motif = motif;
   while (current_motif != NULL)
   {
      motif2 [num_motif] = current_motif;
      num_motif++;
      current_motif = current_motif -> next;
   }
   qsort (motif2, num_motif, sizeof (motif_node *), motif_compare);
   paint = (int **) malloc (num_sequence * sizeof (int *));
   if (errno == ENOMEM)
   {
      printf ("Error: Not enough memory for paint\n");
      exit (-1);
   }
   for (i = 0; i < num_sequence; i++)
   {
      len_sequence = strlen (sequence [i]);
      paint [i] = (int *) malloc (len_sequence * sizeof (int));
      if (errno == ENOMEM)
      {
         printf ("Error: Not enough memory for paint\n");
         exit (-1);
      }
      for (j = 0; j < len_sequence; j++)
         paint [i][j] = 0;
   }
   printf ("%d\n", num_motif);
   for (i = 0; i < num_motif; i++)
   {
      found = 0;
      for (j = 0; j < num_sequence; j++)
      {
         len_sequence = strlen (sequence [j]);
         for (k = 0; k < len_sequence - motif2 [i] -> len_tuple + 1; k++)
         {
            for (l = 0; l < num_strand; l++)
            {
               found2 = 0;
               for (m = 0; m < motif2 [i] -> len_tuple; m++)
               {
                  if (nucleotide (motif2 [i] -> str [m]) >= 0)
                  {
                     if (l == 0)
                        ch = sequence [j][k + m];
                     else
                        ch = complement (sequence [j][k + motif2 [i]
                           -> len_tuple - m - 1]);
                     if (tolower (ch) != motif2 [i] -> str [m])
                     {
                        found2 = 1;
                        break;
                     }
                  }
               }
               if (!found2)
               {
                  for (m = 0; m < motif2 [i] -> len_tuple; m++)
                  {
                     if (paint [j][k + m])
                     {
                        found = 1;
                        break;
                     }
                  }
                  if (found)
                     break;
               }
            }
            if (found)
               break;
         }
         if (found)
            break;
      }
      for (j = 0; j < num_sequence; j++)
      {
         len_sequence = strlen (sequence [j]);
         for (k = 0; k < len_sequence - motif2 [i] -> len_tuple + 1; k++)
         {
            for (l = 0; l < num_strand; l++)
            {
               found2 = 0;
               for (m = 0; m < motif2 [i] -> len_tuple; m++)
               {
                  if (nucleotide (motif2 [i] -> str [m]) >= 0)
                  {
                     if (l == 0)
                        ch = sequence [j][k + m];
                     else
                        ch = complement (sequence [j][k + motif2 [i]
                           -> len_tuple - m - 1]);
                     if (tolower (ch) != motif2 [i] -> str [m])
                     {
                        found2 = 1;
                        break;
                     }
                  }
               }
               if (!found2)
               {
                  for (m = 0; m < motif2 [i] -> len_tuple; m++)
                  {
                     if (l == 0)
                        ch = sequence [j][k + m];
                     else
                        ch = complement (sequence [j][k + motif2 [i]
                           -> len_tuple - m - 1]);
                     printf ("%c", ch);
                     if (!found)
                        paint [j][k + m] = 1;
                  }
                  if (sequence_name [j][0] != ' ')
                     printf (" ");
                  printf ("%s/", sequence_name [j]);
                  if (l > 0)
                     printf ("-");
                  printf ("%d\n", k + 1);
               }
            }
         }
      }
      for (j = 0; j < motif2 [i] -> len_tuple; j++)
      {
         if (nucleotide (motif2 [i] -> str [j]) >= 0)
            printf ("%c", motif2 [i] -> str [j]);
         else
            printf ("-");
      }
      printf (" %d %.2Le", motif2 [i] -> num_word, motif2 [i] -> prob);
      if (found)
         printf ("-");
      printf ("\n");
   }
   free (motif2);
   for (i = 0; i < num_sequence; i++)
      free (paint [i]);
   free (paint);
}

main (int argc, char **argv)
{
   FILE *file;
   int i, j, max_len_tuple, max_len_freq, num_strand, max_num_word,
      num_sequence, *pos;
   char *str, **sequence, **sequence_name;
   freq_node **freq, *current_freq;
   motif_node *motif, *current_motif;

   if ((argc < 2) || (argc > 6))
   {
      printf ("Usage: posmotif file [max_len_tuple] [max_len_freq] [num_strand] [max_num_word]\n");
      exit (-1);
   }
   if (argc <= 2)
      max_len_tuple = MAX_LEN_TUPLE;
   else
   {
      max_len_tuple = atoi (argv [2]);
      if (max_len_tuple < 1)
         max_len_tuple = MAX_LEN_TUPLE;
   }
   if (argc <= 3)
      max_len_freq = MAX_LEN_FREQ;
   else
   {
      max_len_freq = atoi (argv [3]);
      if (max_len_freq < 1)
         max_len_freq = MAX_LEN_FREQ;
   }
   if (argc <= 4)
      num_strand = NUM_STRAND;
   else
   {
      num_strand = atoi (argv [4]);
      if ((num_strand < 1) || (num_strand > 2))
         num_strand = NUM_STRAND;
   }
   if (argc <= 5)
      max_num_word = MAX_NUM_WORD;
   else
   {
      max_num_word = atoi (argv [5]);
      if (max_num_word < 1)
         max_num_word = MAX_NUM_WORD;
   }
   str = fmakeword (stdin, EOF, NULL);
   sequence = read_fasta (str, &sequence_name, &num_sequence);
   free (str);
   freq = (freq_node **)
      malloc ((((max_len_tuple > 1) && (max_len_freq == 1)) ? 1 :
         max (0, max_len_tuple - max_len_freq)) * sizeof (freq_node *));
   if (errno == ENOMEM)
   {
      printf ("Error: Not enough memory for freq\n");
      exit (-1);
   }
   motif = NULL;
   file = fopen (argv [1], "r");
   if (file == NULL)
   {
      printf ("Error: Cannot open file %s\n", argv [1]);
      exit (-1);
   }
   pos = (int *) malloc (max_len_tuple * sizeof (int));
   if (errno == ENOMEM)
   {
      printf ("Error: Not enough memory for pos\n");
      exit (-1);
   }
   pos [0] = 0;
   for (i = 0; i < max_len_tuple; i++)
   {
      for (j = min (1, i); j <= i; j++)
      {
         if ((i == max_len_tuple - 1) || (j != max_len_freq - 1))
            current_freq = NULL;
         else
         {
            current_freq = (freq_node *) malloc (sizeof (freq_node));
            if (errno == ENOMEM)
            {
               printf ("Error: Not enough memory for current_freq\n");
               exit (-1);
            }
            current_freq -> next = NULL;
            freq [i - max_len_freq + 1] = current_freq;
         }
         pos [j] = i;
         compute_motif (max_len_freq, num_strand, max_num_word, num_sequence,
            sequence, freq, current_freq, file, i + 1, j + 1, 1, pos, 1,
            &motif);
      }
   }
   for (i = 0; i < (((max_len_tuple > 1) && (max_len_freq == 1)) ? 1
           : max_len_tuple - max_len_freq); i++)
      free_freq (freq [i]);
   free (freq);
   free (pos);
   if (fclose (file) == EOF)
   {
      printf ("Error: Cannot close file %s\n", argv [1]);
      exit (-1);
   }
   print_motif (num_strand, num_sequence, sequence, sequence_name, motif);
   for (i = 0; i < num_sequence; i++)
      free (sequence [i]);
   free (sequence);
   for (i = 0; i < num_sequence; i++)
      free (sequence_name [i]);
   free (sequence_name);
   while (motif != NULL)
   {
      current_motif = motif;
      motif = motif -> next;
      free (current_motif -> str);
      free (current_motif);
   }
}
