# run as: perl postprocess2.pl < inputfile > outputfile


#! /usr/local/bin/perl -w
use strict;
use POSIX qw(ceil floor);

my @Motifs=();
my %motifhash=();
my ($motifhash_ref, $motifseq_ref)=read_em();
my %motifhash=%$motifhash_ref;
my %motifseqhash=%$motifseq_ref;



my $num=$#Motifs+1;
print "$num \n";


my %CostHash=();

foreach my $motif (@Motifs)
{
  foreach my $motif2(@Motifs)
  {
    if($motif ne $motif2)
    {
      $CostHash{$motif}{$motif2}=alignment($motif,$motif2);
      my $revc_motif2=revcomp($motif2);
      my $score1=alignment($motif,$revc_motif2); 
      if($score1<$CostHash{$motif}{$motif2})
      {
         $CostHash{$motif}{$motif2}=$score1;
      } 
      #print "For $motif and $motif2, the score is $CostHash{$motif}{$motif2} \n";
    }
  }
}



#print "Done with calculating the cost. \n";

my %neighbors=();
my %CheckNeighbors=();
foreach my $motif (@Motifs)
{
   $neighbors{$motif}=0;
   my $len=length($motif);
   my $max=$len;
   foreach my $motif2 (@Motifs)
   {
     my $cost=$CostHash{$motif}{$motif2};
     my $len2=length($motif2);
     $CheckNeighbors{$motif}{$motif2}=0;
     if($max>$len2) {$max=$len2;}
     my $result=index($motif,$motif2);
     if($result==-1) { $result=index($motif2,$motif); }
     if(($motif ne $motif2) && ($result!=-1) )  {
         $neighbors{$motif}++;
         $CheckNeighbors{$motif}{$motif2}=1;
     }
     elsif(($motif ne $motif2)&&($cost<0.2*$max)&&(abs($len2-$len)<0.2*$max))
     {
        $neighbors{$motif}++;
        $CheckNeighbors{$motif}{$motif2}=1;
     }
     else { ; }
   }
}


my @sorted0 =sort mysort  keys %neighbors;
sub mysort {
   my $new=$motifhash{$a};                
   (my $first,my $second_a)=split(/ X /,$new);
   $new=$motifhash{$b};
   (my $first,my $second_b)=split(/ X /,$new);
   $neighbors{$b} <=> $neighbors{$a} ||$second_a <=> $second_b || length($a) <=>length($b);
}

my %rank=();
$num=0;
for (my $i=0;$i<$#sorted0+1;$i++){
  my $motif=$sorted0[$i];
  my $new=$motifhash{$motif};
  (my $first,my $secondx)=split(/ X /,$new);
  $rank{$motif}=ceil(($num+$secondx)/2);
  #$num++;
}

my @sorted =sort mysort2  keys %rank;
sub mysort2 {
   my $new=$motifhash{$a};
   (my $first,my $second_a)=split(/ X /,$new);
   $new=$motifhash{$b};
   (my $first,my $second_b)=split(/ X /,$new);
   $rank{$a} <=> $rank{$b} ||$second_a <=> $second_b || length($a) <=>length($b);
}



my %markedmotifs=();
foreach my $motif2 (@sorted){
  $markedmotifs{$motif2}=0;
}

$num=0;
for (my $i=0;$i<$#sorted+1;$i++){
  my $motif=$sorted[$i];
  if($markedmotifs{$motif}!=1){
    $markedmotifs{$motif}==1; 
    my $len=length($motif);
    my $max=$len;
    my %instances=();
    my $new=$motifhash{$motif};
    (my $first,my $secondx)=split(/ X /,$new);
    my @insts=split(/ o /,$first);
    #print "Itself: $motif\n";
    foreach my $inst (@insts)
    {
       (my $seq,my $pos,my $str)=split(/,/,$inst);
        $instances{$inst}=$pos;
        #print   "$inst\n";
    }
    my $lcmotif=lc($motif);
    #print "Motif $lcmotif has $neighbors{$motif} neighbors \n";
     foreach my $motif2 (@Motifs){
       if($CheckNeighbors{$motif}{$motif2}==1 ) {
           my $num_shared=0;
           foreach my $motif3 (@Motifs){
             if($motif3 ne $motif && $motif3 ne $motif2 && $CheckNeighbors{$motif}{$motif3}==1 && $CheckNeighbors{$
motif2}{$motif3}==1)
             {
               $num_shared++;
             }
           }
           if($num_shared>0.5*$neighbors{$motif}){
               $markedmotifs{$motif2}=1;
           }
           my $new=$motifhash{$motif2};
           (my $first,my $second)=split(/ X /,$new);
           my @insts=split(/ o /,$first);
           #print "Neighbor: $motif2 shared with $motif $num_shared \n";
           foreach my $inst (@insts)
           {
             (my $seq,my $pos,my $str)=split(/,/,$inst);
             $instances{$inst}=$pos;      
             #print   "$inst\n";
           }
       }
     }


     #print "Merged Instances:\n";
    my @sorted_insts=sort { $instances{$b} <=> $instances{$a} ||length($b) <=> length($a) } keys %instances;    
    merge_instances(\@sorted_insts,$motif);
    $lcmotif=~tr/[n]/[\-]/;
    print "$lcmotif $secondx $neighbors{$motif} $num\n";
    $num++;
  }
}


# subroutine to get the reverse complement string of a string
sub revcomp {
  my $motif=$_[0];
  my @motifc=split(//,$motif);
  my $revcomp_str="";
  foreach my $c (@motifc)
  {
     my $copy=$revcomp_str;
     my $comp_c="";
     
     if($c eq 'A') { $comp_c='T';}
     elsif ($c eq 'T') { $comp_c='A';}
     elsif ($c eq 'C') { $comp_c='G';}
     elsif ($c eq 'G') { $comp_c='C';}
     elsif ($c eq 'Y') { $comp_c='R';}
     elsif ($c eq 'R') { $comp_c='Y';}
     elsif ($c eq 'S') { $comp_c='S';}
     elsif ($c eq 'W') { $comp_c='W';}
     elsif ($c eq 'K') { $comp_c='M';}
     elsif ($c eq 'M') { $comp_c='K';}
     elsif ($c eq 'B') { $comp_c='V';}
     elsif ($c eq 'D') { $comp_c='H';}
     elsif ($c eq 'H') { $comp_c='D';}
     elsif ($c eq 'V') { $comp_c='B';}
     elsif ($c eq 'N') { $comp_c='N';}
     else { print "There is error in the IUPAC Degenerate char $c \n ";}
   
     $revcomp_str=$comp_c.$copy; 
   }  
   return $revcomp_str;
}



# subroutine to merge instances inside the cluster
#
sub merge_instances {
  my (@sortedInstances)=@{$_[0]};
  my $motif=$_[1];
  my %InstanceHash=();
  foreach my $line (@sortedInstances) {
   (my $seq,my $pos,my $instance)=split(/,/,$line);       
    my $end=$pos+length($instance);
    my $pos0=$pos;
    my $flag=0;
    for( $pos0=$pos;$pos0<=$end;$pos0++){
       if(exists $InstanceHash{$seq}{$pos0}){
           $flag=1;
           my $oldString=$InstanceHash{$seq}{$pos0};
           my $newString=substr($instance,0,$pos0-$pos);
           if($pos0==$pos){
               if(length($newString)<length($oldString)){
                  $newString=$oldString;
               }
               $InstanceHash{$seq}{$pos0}=$newString;
          }
          else {
               if((length($instance)-length($newString))<length($oldString)){
                   $newString=$newString.$oldString;
                   delete $InstanceHash{$seq}{$pos0};
                   $InstanceHash{$seq}{$pos}=$newString;
               }
               else {
                   $newString=$instance;
                   delete $InstanceHash{$seq}{$pos0};
                   $InstanceHash{$seq}{$pos}=$newString;
               }
         }
     }
   }
    if($flag==0){
      $InstanceHash{$seq}{$pos}=$instance;
    }
 }  
 
 my @sortedseqs=sort {
          $motifseqhash{$motif}{$a} <=> $motifseqhash{$motif}{$b} } keys %InstanceHash;
 
 foreach my $seq (@sortedseqs){
    my @sortedpos=sort { $a <=> $b } keys %{ $InstanceHash{$seq} };
    foreach my $pos (@sortedpos){
     print "$InstanceHash{$seq}{$pos} $seq\/$pos \n";
   }
 }
 return (%InstanceHash);
}


# sub routine to return minimum cost alignment between two strings
sub alignment{
  my $A=$_[0];
  my $B=$_[1];
  my $n=length($A);
  my $m=length($B);
  #print "Alignment between $A and $B \n";

  my $min_score=$n+$m;
  my $flag=0;
  if($n<$m)
  {
    $A=$_[1];
    $B=$_[0];
    $n=length($A);
    $m=length($B);
    $flag=1;
  }
  for(my $i=0;$i<$n+$m-1;$i++)
  {
    #i is the start position of B
    my $score=0;
    my $L=$i;
    if($i<$m-1) # B in the left-side case
    {
      for(my $j=0;$j<$m-1-$i;$j++)
      {
         my $s=substr($B,$j,1);
         if(uc($s) ne 'N') { $score+=0.7;}
      }
      for(my $j=0;$j<$i+1;$j++)
      {
        my $x=substr($A,$j,1);
        my $s=substr($B,$m-$i-1+$j,1);
         #print " the x and s is $x and $s  \n";
        if($flag==0)
         {$score+=subset($x,$s);}
        else
         {$score+=subset($s,$x);}
      }
      for(my $j=0;$j<$n-1-$i;$j++)
      {
         my $x=substr($A,$L+$j,1);
         if(uc($x) ne 'N') { $score+=0.7; }
      }
    }
 elsif($i>$n-1) # B in the right-side case
    {
      my $L=$n+$m-$i-1;
      #print " the L is $L \n";
      for(my $j=0;$j<$i-$m+1;$j++)
      {
         my $x=substr($A,$j,1);
         if(uc($x) ne 'N') {
         $score+=0.7;}
      }
      for(my $j=0;$j<$L;$j++)
      {
       my $x=substr($A,$i-$m+1+$j,1);
       my $s=substr($B,$j,1);
       #print " the x and s is $x and $s  \n";
       if($flag==1)
        {$score+=subset($s,$x);}
        else {$score+=subset($x,$s);}
      }
      for(my $j=0;$j<$i-$n+1;$j++)
      {
        my $s=substr($B,$L+$j,1);
        if(uc($s) ne 'N') {
         $score+=0.7;}
      }
    }
 else # B in the middle case
    {
      my $L=$m;
      for(my $j=0;$j<$i-$m+1;$j++)
      {
         my $x=substr($A,$j,1);
         if(uc($x) ne 'N') {
         $score+=0.7;}
      }

      for(my $j=0;$j<$m;$j++)
      {
        my $x=substr($A,$i-$m+1+$j,1);
        my $s=substr($B,$j,1);
        #print " the x and s is $x and $s  \n";
        if($flag==0){$score+=subset($x,$s);}
        else {$score+=subset($s,$x);}
      }
      for(my $j=0;$j<$n-$i-1;$j++)
      {
         my $x=substr($A,$i+1+$j,1);
         if(uc($x) ne 'N') {
          $score+=0.7;}
      }
   }
   #print " for i=$i the score is $score \n";
   if($min_score>$score) { $min_score=$score;}
  }
  return $min_score;
}



# subroutine to determine is s is a subset of x
sub subset {
 my $x=$_[0];
 my $s=$_[1];
 if($x eq 'N' )
 {
   if($s eq $x) { return 0; }
   else { return 0.5;}
 }
 elsif($x eq 'A')
 {
  if($x eq $s) { return 0;}
  elsif($s=~/(C|G|T|Y|S|K|B)/) { return 1;}
  else { return 0.5; }
 }
 elsif($x eq 'C')
 {
  if($x eq $s) { return 0;}
  elsif($s=~/(A|G|T|R|W|K|D)/) { return 1;}
  else { return 0.5; }
 }
  elsif($x eq 'G')
 {
  if($x eq $s) { return 0;}
  elsif($s=~/(A|C|T|Y|W|M|H)/) { return 1;}
  else { return 0.5; }
 }
  elsif($x eq 'T')
 {
  if($x eq $s) { return 0;}
  elsif($s=~/(A|C|G|R|S|M|V)/) { return 1;}
  else { return 0.5; }
 }
 else { printf(" Error in the $x and $s \n"); exit; }
}


sub read_em {
  my %instances=();
  my $num=1;
  my $flag=0;
  my %motifhash=();
  my %motifseqhash=();
  my %seqhash=();
  my $seqnum=0;
  while(my $line=<STDIN>){
   # print " The Line is $line \n";
     if($line=~/(^[ACGTNacgtn]+)(\s+)(\w+)(\/)(\d+)/){
        # print " The instance line0: $line \n";
         my $instance=$1;                                                                             
         my $seq=$3;                                                                                  
         my $pos=$5;                                                                                  
         my $report="$seq,$pos,$instance";                                                            
         $instances{$report}=$pos ;      
         $seqhash{$seq}=$seqnum;
         $seqnum++;
      }
     elsif($line=~/(^[ACGTNacgtn]+)(\s+)(\S+)(\s+.*)(\/)(\d+)/){
         #print " The instance line1: $line \n"; 
         my $instance=$1;
         my $seq=$3;
         my $pos=$6;
         my $report="$seq,$pos,$instance"; 
         $instances{$report}=$pos ;
         $seqhash{$seq}=$seqnum; 
         $seqnum++;
      }
     elsif($line=~/([ACGTacgt-]+)(\s*)(\d+\.\d+)(e\-\d+)(\s+)(\d+)(\s*.*)/){
       #print " The motif line: $line \n";
       my $motif=$1;
       my $score1=$3;                                                                                                                             
       my $score2=$4;  
       my $rank=$6;
       if($flag!=1) {       
              my @sortedInstances=sort { $instances{$b} <=> $instances{$a} } keys %instances;
              my $motifInstances=join(' o ',@sortedInstances);
              my $motifInstances2=$motifInstances." X ".$score1.$score2." X ".$rank;
              $motif=~tr/[acgt\-]/[ACGTN]/;
              $motifhash{$motif}=$motifInstances2;
              push(@Motifs,$motif);
              foreach my $inst (@sortedInstances)
              {
                 my ($seq,$pos,$str)=split(/,/,$inst);
                 $motifseqhash{$motif}{$seq}=$seqhash{$seq};
              }
              %seqhash=();
              $seqnum=0;
              %instances=();
              if($num>100) { last;}
              $num=$num+1;
              
      }
      %instances=();
      %seqhash=();
      $seqnum=0;
      $flag=0; 
   }     
  else { ; }
  }
 return (\%motifhash, \%motifseqhash);
}



