#include <stdio.h>
#include <errno.h>

#define MAX_LEN_TUPLE 20
#define MAX_LEN_FREQ   3
#define NUM_STRAND     1

#define NUM_NUCLEOTIDE 4

#define min(a,b) ((a < b) ? a : b)
#define nucleotide(c) ((c=='a')?0:((c=='c')?1:((c=='g')?2:((c=='t')?3:-1))))

char nucleotide_char [] =
{
   /* 0 */ 'a', /* 1 */ 'c', /* 2 */ 'g', /* 3 */ 't'
};

char *fmakeword(FILE *f, char stop, int *cl) {
    int wsize;
    char *word;
    int ll;

    wsize = 102400;
    ll=0;
    word = (char *) malloc(sizeof(char) * (wsize + 2));

    while(1) {
        word[ll] = (char)fgetc(f);
        if(ll==wsize) {
            word[ll+1] = '\0';
            wsize+=102400;
            word = (char *)realloc(word,sizeof(char)*(wsize+2));
        }
        if (cl != NULL)
           --(*cl);
        if((word[ll] == stop) || (feof(f)) || ((cl != NULL) && (!(*cl)))) {
            if((word[ll] != stop) && (!feof(f)) && ((cl == NULL) || (!(*cl))))
                ll++;
            word[ll] = '\0';
            return word;
        }
        ++ll;
    }
}

char complement (char ch)
{
   if (ch == 'a')
      return 't';
   if (ch == 'c')
      return 'g';
   if (ch == 'g')
      return 'c';
   if (ch == 't')
      return 'a';
   return ch;
}

char **read_fasta (char *str, char ***sequence_name, int *num_sequence)
{
   int i, n, start, end, len_sequence;
   char **sequence;

   n = strlen (str);
   (*num_sequence) = 0;
   i = 0;
   while (i < n)
   {
      while ((i < n) && (str [i] != '>'))
         i++;
      while ((i < n) && (str [i] != '\n'))
         i++;
      while ( (i < n) && (str [i] != '>') && (!islower (str [i]))
           && (!isupper (str [i])) )
         i++;
      if ((i < n) && (str [i] != '>'))
         (*num_sequence)++;
   }
   sequence = (char **) malloc ((*num_sequence) * sizeof (char *));
   if (errno == ENOMEM)
   {
      printf ("Error: Not enough memory for sequence\n");
      exit (-1);
   }
   (*sequence_name) = (char **) malloc ((*num_sequence) * sizeof (char *));
   if (errno == ENOMEM)
   {
      printf ("Error: Not enough memory for sequence_name\n");
      exit (-1);
   }
   (*num_sequence) = 0;
   i = 0;
   while ((i < n) && (str [i] != '>'))
      i++;
   while (i < n)
   {
      start = i + 1;
      while ((i < n) && (str [i] != '\n'))
         i++;
      end = i - 1;
      while ( (i < n) && (str [i] != '>') && (!islower (str [i]))
           && (!isupper (str [i])) )
         i++;
      if ((i < n) && (str [i] != '>'))
      {
         (*sequence_name) [*num_sequence] = (char *)
            malloc ((end - start + 2) * sizeof (char));
         if (errno == ENOMEM)
         {
            printf ("Error: Not enough memory for sequence_name\n");
            exit (-1);
         }
         strncpy ((*sequence_name) [*num_sequence], &str [start],
            end - start + 1);
         (*sequence_name) [*num_sequence][end - start + 1] = '\0';
         start = i;
         len_sequence = 0;
         while ((i < n) && (str [i] != '>'))
         {
            if ((islower (str [i])) || (isupper (str [i])))
               len_sequence++;
            i++;
         }
         sequence [*num_sequence] = (char *)
            malloc ((len_sequence + 1) * sizeof (char));
         if (errno == ENOMEM)
         {
            printf ("Error: Not enough memory for sequence\n");
            exit (-1);
         }
         i = start;
         len_sequence = 0;
         while ((i < n) && (str [i] != '>'))
         {
            if ((islower (str [i])) || (isupper (str [i])))
            {
               sequence [*num_sequence][len_sequence] = tolower (str [i]);
               len_sequence++;
            }
            i++;
         }
         sequence [*num_sequence][len_sequence] = '\0';
         (*num_sequence)++;
      }
   }
   return sequence;
}

void compute_motif (int num_strand, int num_sequence, char **sequence,
   int len_tuple, int len_word, int len_word2, int *pos, int pos2)
{
   int i, j, k, l, index, index2, *num_dist, len_sequence;
   char ch;

   if (len_word <= len_word2 + 1)
   {
      index = 1;
      for (i = 0; i < len_word; i++)
         index *= NUM_NUCLEOTIDE;
      num_dist = (int *) malloc (index * sizeof (int));
      if (errno == ENOMEM)
      {
         printf ("Error: Not enough memory for num_dist\n");
         exit (-1);
      }
      for (i = 0; i < index; i++)
         num_dist [i] = 0;
      for (i = 0; i < num_sequence; i++)
      {
         len_sequence = strlen (sequence [i]);
         for (j = 0; j < len_sequence - len_tuple + 1; j++)
         {
            for (k = 0; k < num_strand; k++)
            {
               index2 = 0;
               for (l = 0; l < len_word; l++)
               {
                  if (k == 0)
                     ch = sequence [i][j + pos [l]];
                  else
                     ch = complement (sequence [i][j + len_tuple - pos [l]
                        - 1]);
                  if (nucleotide (ch) < 0)
                  {
                     index2 = -1;
                     break;
                  }
                  index2 = NUM_NUCLEOTIDE * index2 + nucleotide (ch);
               }
               if (index2 >= 0)
                  num_dist [index2]++;
            }
         }
      }
      for (i = 0; i < index; i++)
         printf ("%d\n", num_dist [i]);
      free (num_dist);
   }
   else
   {
      for (i = pos2; i <= len_tuple - len_word + len_word2; i++)
      {
         pos [len_word2] = i;
         compute_motif (num_strand, num_sequence, sequence, len_tuple, len_word,
            len_word2 + 1, pos, i + 1);
      }
   }
}

main (int argc, char **argv)
{
   int i, j, max_len_tuple, max_len_freq, num_strand, num_sequence, *pos;
   char *str, **sequence, **sequence_name;

   if (argc > 4)
   {
      printf ("Usage: freq0 [max_len_tuple] [max_len_freq] [num_strand]\n");
      exit (-1);
   }
   if (argc <= 1)
      max_len_tuple = MAX_LEN_TUPLE;
   else
   {
      max_len_tuple = atoi (argv [1]);
      if (max_len_tuple < 1)
         max_len_tuple = MAX_LEN_TUPLE;
   }
   if (argc <= 2)
      max_len_freq = MAX_LEN_FREQ;
   else
   {
      max_len_freq = atoi (argv [2]);
      if (max_len_freq < 1)
         max_len_freq = MAX_LEN_FREQ;
   }
   if (argc <= 3)
      num_strand = NUM_STRAND;
   else
   {
      num_strand = atoi (argv [3]);
      if ((num_strand < 1) || (num_strand > 2))
         num_strand = NUM_STRAND;
   }
   str = fmakeword (stdin, EOF, NULL);
   sequence = read_fasta (str, &sequence_name, &num_sequence);
   free (str);
   pos = (int *) malloc (max_len_tuple * sizeof (int));
   if (errno == ENOMEM)
   {
      printf ("Error: Not enough memory for pos\n");
      exit (-1);
   }
   pos [0] = 0;
   for (i = 0; i < max_len_tuple; i++)
   {
      for (j = min (1, i); j < min (max_len_freq, i + 1); j++)
      {
         pos [j] = i;
         compute_motif (num_strand, num_sequence, sequence, i + 1, j + 1, 1,
            pos, 1);
      }
   }
   for (i = 0; i < num_sequence; i++)
      free (sequence [i]);
   free (sequence);
   for (i = 0; i < num_sequence; i++)
      free (sequence_name [i]);
   free (sequence_name);
   free (pos);
}
